﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.Threading;

namespace reversi
{
    public partial class Reversi : Form
    {
        private enum GameState
        {
            GameOver,			// The game is over (also used for the initial state).
            InMoveAnimation,	// A move has been made and the animation is active.reserve for future Development.
            InPlayerMove,		// Waiting for the user to make a move.
            InComputerMove,		// Waiting for the computer to make a move.
            MoveCompleted		// A move has been completed (including the animation, if active).
        }
        // The game board.
        private Board board;
        private Movement movement;
        private Label[] colLabels, rowLabels;
        private SquareControl[,] squareControls;
        // Game parameters.
        private GameState gameState;
        private int currentColor;
        private int moveNumber;
        // AI parameters.
        private int lookAheadDepth;
        private int forfeitWeight;
        private int frontierWeight;
        private int mobilityWeight;
        private int stabilityWeight;
        // Defines a thread for running the computer move look ahead.
        private Thread calculateComputerMoveThread;
        private struct ComputerMove
        {
            public int row;
            public int col;
            public int rank;

            public ComputerMove(int row, int col)
            {
                this.row = row;
                this.col = col;
                this.rank = 0;
            }
        }
        // Defines the maximum move rank value (used for ranking an end game).
        private static int maxRank = System.Int32.MaxValue - 64;
        // For converting column numbers to letters and vice versa.
        /// <summary>
        /// </summary>
        private static String alpha = "ABCDEFGH";
        // Used to track which player made the last move.
        // Defines a thread for running the computer move look ahead.
        
        public Reversi()
        {
            InitializeComponent();

            // Create the game board.
            this.board = new Board();
            this.movement = new Movement(this.board);
            // Create the controls for each square, add them to the squares
            // panel and set up event handling.
			this.squareControls = new SquareControl[8, 8];
			int i, j;
			for (i = 0; i < 8; i++)
				for (j = 0; j < 8; j++)
				{
					// Create it.
					this.squareControls[i, j] = new SquareControl(i, j);
					// Position it.
					this.squareControls[i, j].Left = j * this.squareControls[i, j].Width;
					this.squareControls[i, j].Top  = i * this.squareControls[i, j].Height;
					// Add it.
					this.squaresPanel.Controls.Add(this.squareControls[i, j]);
					// Set up event handling for it.
					this.squareControls[i, j].MouseMove  += new MouseEventHandler(this.SquareControl_MouseMove);
					this.squareControls[i, j].MouseLeave += new EventHandler(this.SquareControl_MouseLeave);
					this.squareControls[i, j].Click      += new EventHandler(this.SquareControl_Click);
				}

			    // Create the column and row labels.
			    this.colLabels = new Label[8];
			    for (i = 0; i < 8; i++)
			    {
				// Create a column label.
				this.colLabels[i] = new Label();

				// Set its display properties.
				this.colLabels[i].Text = Reversi.alpha.Substring(i, 1);
				this.colLabels[i].BackColor = this.cornerLabel.BackColor;
				this.colLabels[i].ForeColor = this.cornerLabel.ForeColor;
				this.colLabels[i].TextAlign = ContentAlignment.MiddleCenter;

				// Set its size and position.
				this.colLabels[i].Width = this.squareControls[0, 0].Width;
				this.colLabels[i].Height = this.cornerLabel.Height;
				this.colLabels[i].Left = this.cornerLabel.Width + i * this.colLabels[0].Width;
				this.colLabels[i].Top = 0;

				// Add it.
				this.boardPanel.Controls.Add(this.colLabels[i]);
			}
			this.rowLabels = new Label[8];
			for (i = 0; i < 8; i++)
			{
				// Create a row label.
				this.rowLabels[i] = new Label();

				// Set its display properties.
				this.rowLabels[i].Text      = (i + 1).ToString();
				this.rowLabels[i].BackColor = this.cornerLabel.BackColor;
				this.rowLabels[i].ForeColor = this.cornerLabel.ForeColor;
				this.rowLabels[i].TextAlign = ContentAlignment.MiddleCenter;

				// Set its size and position.
				this.rowLabels[i].Width  = this.cornerLabel.Height;
				this.rowLabels[i].Height = this.squareControls[0, 0].Height;
				this.rowLabels[i].Left   = 0;
             
                this.rowLabels[i].Top    = this.cornerLabel.Height + i * this.rowLabels[0].Width*2;
             

				// Add it.
				this.boardPanel.Controls.Add(this.rowLabels[i]);
			}
         
            // Initialize the game state.
            this.gameState = Reversi.GameState.GameOver;
         }
        // ===================================================================
        // This code handles game play.
        // ===================================================================

        //
        // Starts a new game.
        //
        private void StartGame()
        {
            this.movement.SetForNewGame();
            this.UpdateBoardDisplay();
            this.currentColor =  Board.Black;
            // Start the first turn.
            this.StartTurn();
        }
        //
        // Sets up for the current player to make a move or ends the game if
        // neither player can make a valid move.
        //
        private void StartTurn()
        {
            // If the current player cannot make a valid move, forfeit the turn.
           if(this.movement.GetValidMoveCount(this.currentColor)==0)
            {
                // Switch back to the other player.
                this.currentColor *= -1;

                // If the original player cannot make a valid move either, the game is over.
                if (this.movement.GetValidMoveCount(this.currentColor)==0)
                {
                 
                    return;
                }
            }
            // If the current color is under computer control, set up for a
            // computer move.
            if (this.IsComputerPlayer(this.currentColor))
            {
                // Set the game state.
                this.gameState = Reversi.GameState.InComputerMove;

                // Check if computer play is currently suspended.
            
                    // Start a separate thread to perform the computer's move.
                    this.calculateComputerMoveThread = new Thread(new ThreadStart(this.CalculateComputerMove));
                    this.calculateComputerMoveThread.IsBackground = true;
                    this.calculateComputerMoveThread.Priority = System.Threading.ThreadPriority.Lowest;
                    this.calculateComputerMoveThread.Name = "Calculate Computer Move";

                    this.calculateComputerMoveThread.Start();

                
            }

            // Otherwise, set up for a user move.
            else
            {
                // Set the game state.
                this.gameState = Reversi.GameState.InPlayerMove;
            }

      
        }
        //
        // Determines if a given color is being played by the computer.
        //
        private bool IsComputerPlayer(int color)
        {
            if (color == Board.White) return true;
            else return false;
        }
        //
        // Makes a move for the current player.
        //
        private void MakeMove(int row, int col)
        {   
            // Bump the move number.
            this.moveNumber++;
           // Make the move on the board.
            this.movement.MakeMove(this.currentColor, row, col);

            // Update the display to reflect the board changes.
            this.UpdateBoardDisplay();
            // Set the game state.
            this.gameState = Reversi.GameState.MoveCompleted;

            // Switch players and start the next turn.
            this.currentColor *= -1;
            this.StartTurn();
        }
        // ===================================================================
        // Code to handle computer moves.
        // ===================================================================
        //
        // Define delegates for callbacks from the worker thread.
        //
        public delegate void MakeComputerMoveDelegate(int row, int col);
        //
        // Calculates a computer move.
        // Note: Executed in the worker thread.
        //
        private void CalculateComputerMove()
        {
            #region  Sets the AI parameters based on the current difficulty setting.

            this.forfeitWeight = 2;
            this.frontierWeight = 1;
            this.mobilityWeight = 0;
            this.stabilityWeight = 3;
            // Set the look-ahead depth.
            this.lookAheadDepth = 3;

            // Near the end of the game, when there are relatively few moves
            // left, set the look-ahead depth to do an exhaustive search.
            if (this.moveNumber >= 55)
                this.lookAheadDepth = this.board.EmptyCount;
            #endregion
            // Find the best available move.
            // Kick off the look ahead.
            // Initialize the alpha-beta cutoff values.
            int alpha = Reversi.maxRank + 64;//2147483647
            int beta = -alpha;
            ComputerMove move = this.GetBestMove(this.board, this.currentColor, 1, alpha,beta );
          
            // Perform a callback to make the move.
            Object[] args = { move.row, move.col };
            MakeComputerMoveDelegate moveDelegate = new MakeComputerMoveDelegate(this.MakeMove);
            this.BeginInvoke(moveDelegate, args);

        }
        //
      // This function uses look ahead to evaluate all valid moves for a
      // given player color and returns the best move it can find.
      //
        private ComputerMove GetBestMove(Board board, int color, int depth, int alpha, int beta)
      {
          // Initialize the best move.
          ComputerMove bestMove = new ComputerMove(-1, -1);

          bestMove.rank = -color * Reversi.maxRank;

          // Find out how many valid moves we have so we can initialize the
          // mobility score.
          int validMoves = movement.GetValidMoveCount(color);

          // Start at a random position on the board. This way, if two or
          // more moves are equally good, we'll take one of them at random.
          Random random = new Random();
          int rowStart = random.Next(8);
          int colStart = random.Next(8);

          // Check all valid moves.
          int i, j;
          for (i = 0; i < 8; i++)
              for (j = 0; j < 8; j++)
              {
                  // Get the row and column.
                  int row = (rowStart + i) % 8;
                  int col = (colStart + j) % 8;

                  if (movement.IsValidMove(color, row, col))
                  {
                      // Make the move.
                     ComputerMove testMove = new ComputerMove(row, col);
                     Board testBoard = new Board(board);
                     Movement testMovement=new Movement(testBoard);
                     testMovement.MakeMove(color, testMove.row, testMove.col);
                     int score = testBoard.WhiteCount - testBoard.BlackCount;

                      // Check the board.
                      int nextColor = -color;
                      int forfeit = 0;
                      bool isEndGame = false;
                      int opponentValidMoves = testMovement.GetValidMoveCount(nextColor);
                      if (opponentValidMoves == 0)
                      {
                          // The opponent cannot move, count the forfeit.
                          forfeit = color;

                          // Switch back to the original color.
                          nextColor = -nextColor;

                          // If that player cannot make a move either, the
                          // game is over.
                          if (testMovement.GetValidMoveCount(nextColor)==0)
                              isEndGame = true;
                      }

                      // If we reached the end of the look ahead (end game or
                      // max depth), evaluate the board and set the move
                      // rank.
                      if (isEndGame || depth == this.lookAheadDepth)
                      {
                          // For an end game, max the ranking and add on the
                          // final score.
                          if (isEndGame)
                          {
                              // Negative value for black win.
                              if (score < 0)
                                  testMove.rank = -Reversi.maxRank + score;

                              // Positive value for white win.
                              else if (score > 0)
                                  testMove.rank = Reversi.maxRank + score;

                              // Zero for a draw.
                              else
                                  testMove.rank = 0;
                          }

                          // It's not an end game so calculate the move rank.
                          else
                              testMove.rank =
                                  this.forfeitWeight * forfeit +
                                  this.frontierWeight * (testBoard.BlackFrontierCount - testBoard.WhiteFrontierCount) +
                                  this.mobilityWeight * color * (validMoves - opponentValidMoves) +
                                  this.stabilityWeight * (testBoard.WhiteSafeCount - testBoard.BlackSafeCount) +
                                                         score;
                      }

                      // Otherwise, perform a look ahead.
                      else
                      {
                          ComputerMove nextMove = this.GetBestMove(testBoard, nextColor, depth + 1, alpha, beta);

                          // Pull up the rank.
                          testMove.rank = nextMove.rank;

                          // Forfeits are cumulative, so if the move did not
                          // result in an end game, add any current forfeit
                          // value to the rank.
                          if (forfeit != 0 && Math.Abs(testMove.rank) < Reversi.maxRank)
                              testMove.rank += forfeitWeight * forfeit;

                          // Adjust the alpha and beta values, if necessary.
                          //computer is white
                          if (color == Board.White && testMove.rank > beta)
                              beta = testMove.rank;
                          if (color == Board.Black && testMove.rank < alpha)
                              alpha = testMove.rank;
                      }

                      // Perform a cutoff if the rank is outside tha alpha-beta range.
                      if (color == Board.White && testMove.rank > alpha)
                      {
                          testMove.rank = alpha;
                          return testMove;
                      }
                      if (color == Board.Black && testMove.rank < beta)
                      {
                          testMove.rank = beta;
                          return testMove;
                      }

                      // If this is the first move tested, assume it is the
                      // best for now.
                      if (bestMove.row < 0)
                          bestMove = testMove;

                      // Otherwise, compare the test move to the current
                      // best move and take the one that is better for this
                      // color.
                      else if (color * testMove.rank > color * bestMove.rank)
                          bestMove = testMove;
                  }
              }

          // Return the best move found.
          return bestMove;
      }

        private void UpdateBoardDisplay()
        {
            // Map the current game board to the square controls.
            SquareControl squareControl;
            int i, j;
            for (i = 0; i < 8; i++)
                for (j = 0; j < 8; j++)
                {
                    
                    squareControl = (SquareControl)this.squaresPanel.Controls[i * 8 + j];
                    squareControl.Contents = this.movement.board.squares[i, j];
                  
                  
                }

            // Redraw the board.

            this.squaresPanel.Refresh();
        }

        //
   // Handles a "New Game" click.
   //
        private void Reversi_Load(object sender, EventArgs e)
    {	
       // Start a new game.
       this.StartGame();
    }
        // ===================================================================
        // Event handlers for the square controls.
        // ===================================================================

        //
        // Handles a mouse move on a board square.
        //
        private void SquareControl_MouseMove(object sender, MouseEventArgs e)
        {
            // Check the game state to ensure that it is the user's turn.
            if (this.gameState != Reversi.GameState.InPlayerMove)
                return;

            SquareControl squareControl = (SquareControl)sender;

            // If the square is a valid move for the current player,
            // indicate it.
            if (this.movement.IsValidMove(this.currentColor, squareControl.row, squareControl.col))
            {
                // Change the cursor.
                squareControl.Cursor = System.Windows.Forms.Cursors.Hand;
                    
             }

               
            
        }

        //
        // Handles a mouse leave on a board square.
        //
        private void SquareControl_MouseLeave(object sender, System.EventArgs e)
        {
            SquareControl squareControl = (SquareControl)sender;
            // If the move is being previewed, clear all affected squares.
            // Restore the cursor.
            squareControl.Cursor = System.Windows.Forms.Cursors.Default;
        }

        //
        // Handles a click on a board square.
        //
        private void SquareControl_Click(object sender, System.EventArgs e)
        {
            // Check the game state to ensure it's the user's turn.
            if (this.gameState != Reversi.GameState.InPlayerMove)
                return;

            SquareControl squareControl = (SquareControl)sender;

            // If the move is valid, make it.
            if (this.movement.IsValidMove(this.currentColor, squareControl.row, squareControl.col))
            {
                // Restore the cursor.
                squareControl.Cursor = System.Windows.Forms.Cursors.Default;

                // Make the move.
                this.MakeMove(squareControl.row, squareControl.col);
            }
        }

    }
}
